use io;

/* Read instance data */
function input() {

    local usage = "Usage : localsolver location_routing_problem.lsp "
            + "inFileName=inputFile [solFileName=outputFile] [lsTimeLimit=timeLimit]";
    if (inFileName == nil) throw usage;

    if (lsTimeLimit == nil) lsTimeLimit=20;

    nbDepotCopies = 20;

    readInputMultiTripVrp();
    
    nbTotalLocations = nbCustomers + nbDepots*nbDepotCopies;
    
    computeDistances();

    maxDist = 400;

    truckCapacity = truckCapacity/2;

    nbTrucks = 3;

}

function model(){

    // add an extra fictive truck
    visitOrder[0...nbTrucks+1] <- list(nbTotalLocations);

    // Constraint on the fictive truck (can only "visit" depots)
    for[i in 0...nbCustomers] constraint !contains(visitOrder[nbTrucks],i);

    // All clients are visited
    constraint partition(visitOrder);

    // Capacity constraints
    for[k in 0...nbTrucks]{
        local sequence <- visitOrder[k];
        local c <- count(sequence);

        // A truck is used if it visits at least one customer
        truckUsed[k] <- c > 0;

        // Distance traveled by truck k
        routeDistances[k] <- sum(1...c, i => distanceMatrix[sequence[i - 1]][sequence[i]])
            + (truckUsed[k] ?
                (distanceMatrix[sequence[0]][nbCustomers] + distanceMatrix[sequence[c - 1]][nbCustomers]) :
                0);

        // The quantity needed in each route must not exceed the truck capacity
        // at any point in the sequence
        routeQuantity[k] <- array( 
            0...c, 
            (i,prev) => (sequence[i] < nbCustomers) ? prev+demands[sequence[i]]:0,
            0);
        constraint and(0...c, i => routeQuantity[k][i]<=truckCapacity);

        constraint routeDistances[k] <= maxDist;
    }

    totalDistance <- sum[k in 0...nbTrucks](routeDistances[k]);

    minimize totalDistance;
}

function readInputMultiTripVrp() {
    if (inFileName.endsWith(".dat")) readInputDat();
    else throw "Unknown file format";
}

function readInputDat() {
    local inFile = io.openRead(inFileName);

    nbCustomers = inFile.readInt();
    nbDepots = inFile.readInt();

    coordinatesDepots[d in 0...nbDepots][l in 0...2] = inFile.readDouble();
    coordinatesCustomers[c in 0...nbCustomers][l in 0...2] = inFile.readDouble();

    truckCapacity = inFile.readInt();

    // Ignore information on depot capacity on this problem
    depotsCapacity[d in 0...nbDepots] = inFile.readInt(); 

    demands[c in 0...nbCustomers] = inFile.readInt();
}

function computeDist(xi, xj, yi, yj) {
    local exactDist = sqrt(pow((xi - xj), 2) + pow((yi - yj), 2));
    return round(exactDist);
}

function computeDistances() {
    for [i in 0...nbCustomers] {
        distanceMatrix[i][i] = 0;
        for [j in i+1...nbCustomers] {
            local localDistance = computeDist(coordinatesCustomers[i][0], coordinatesCustomers[j][0], coordinatesCustomers[i][1], coordinatesCustomers[j][1]);
            distanceMatrix[j][i] = localDistance;
            distanceMatrix[i][j] = localDistance;
        }
    }

    for [i in 0...nbCustomers] {
        for [d in 0...nbDepots]{
            local localDistance = computeDist(coordinatesDepots[d][0], coordinatesCustomers[i][0], coordinatesDepots[d][1], coordinatesCustomers[i][1]);
            for [c in 0...nbDepotCopies] {
                j = nbCustomers + d*nbDepotCopies + c;
                distanceMatrix[i][j] = localDistance;
                distanceMatrix[j][i] = localDistance;
            }
        }
    }

    for [i in nbCustomers...nbTotalLocations] {
        for [j in nbCustomers...nbTotalLocations] {
            distanceMatrix[i][j] = 100000;
        }
    }
}

function output() {
    if (solFileName == nil) return;
    local outfile = io.openWrite(solFileName);
    outfile.println("File name: " + inFileName +"; totalDistance = " + totalDistance.value);
    for [k in 0...nbTrucks] {
        if(truckUsed[k].value){
            outfile.print("Truck " + k + " : ");
            for [customer in visitOrder[k].value]{
                outfile.print(customer<nbCustomers ? customer:-(floor((customer-nbCustomers)/nbDepotCopies) + 1), " ");
            }
            outfile.println();
        }
    }
}